(** Consistency of the numerical scheme *)

Require Export Fourier.
Require Export one_d_wave_equation.
Open Local Scope nat_scope.
Open Local Scope R_scope.

Section Consistency. 
Variable c:R.
Variable usol : R * R -> R.
Variable u0 u1: R->R.
Variable f : R*R -> R.

Variable u0h u1h: Z-> R.
Variable fh: Z -> nat -> R.
Variable A B eta:R.

Hypothesis usol_is_sol: is_solution c u0 u1 f usol.
Hypothesis u0h_is_u0: forall dX j, u0h j = u0 (fst (Xjk dX j 0)).
Hypothesis u1h_is_u1: forall dX j, u1h j = u1 (fst (Xjk dX j 1)).
Hypothesis fh_is_f: forall dX j k, fh j k = f (Xjk dX j k).

Hypothesis f_is_zero_at_beginning: forall x, f (x,0) = 0.

Hypothesis u0_zero: forall x, (x < A)  \/ (B < x) 
   -> u0 x = 0.
Hypothesis u1_zero: forall x, (x < A) \/ (B < x) 
   -> u1 x = 0.
Hypothesis f_zero: forall x t, (x < A-c*t)  \/ (B +c*t< x) 
   -> f (x,t) = 0.
Hypothesis c_pos: 0 < c.
Hypothesis A_le_B: A <= B.

Hypothesis eta_pos: 0 < eta.

Notation OupsCFL_l := (OuP (fun dX => 0 < fst dX /\ 0 < snd dX /\ CFL_condition_l c eta dX)).

(** Finite support of various functions *)

Lemma FS_truncation_error_aux: forall dX k, (0 < fst dX) -> (0 < snd dX) ->
 (forall j, (j < Int_part (A/fst dX) - Z_of_nat k *(up (c*snd dX/ fst dX))
         \/ Int_part (B/fst dX) + Z_of_nat k*(up (c*snd dX/ fst dX)) < j)%Z -> 
   truncation_error c usol u0h u1h fh dX j k = 0).
intros dX k L1 L2 j H.
case_eq k; intros.
rewrite H0 in H;simpl;simpl in H.
rewrite (u0h_is_u0 dX).
rewrite u0_zero.
2: apply zeroness_unh_aux1; case H;auto with zarith.
unfold ubar, discretization.
rewrite (zeroness_usol c u0 u1 f usol A B); trivial.
auto with real.
apply zeroness_unh_aux2; trivial.
assert  (V:(1 <= (up (c * snd dX / fst dX)))%Z).
assert (0 < up (c * snd dX / fst dX))%Z.
apply up_gt_zero.
repeat apply Rmult_lt_0_compat; auto with real.
omega.
case_eq n; intros.
simpl.
unfold Ah, ubar, discretization.
repeat rewrite (zeroness_usol c u0 u1 f usol A B); auto.
rewrite (u1h_is_u1 dX).
rewrite u1_zero.
unfold Rdiv; ring.
apply zeroness_unh_aux1; auto.
rewrite H1 in H0; rewrite H0 in H; simpl (Z_of_nat 1) in H.
case H; intros; auto with zarith.
apply zeroness_unh_aux2; trivial.
rewrite H1 in H0; rewrite H0 in H; simpl (Z_of_nat 1) in H.
simpl; case H; auto with zarith.
apply zeroness_unh_aux2; trivial.
rewrite H1 in H0; rewrite H0 in H; simpl (Z_of_nat 1) in H.
simpl; case H; auto with zarith.
apply zeroness_unh_aux2; trivial.
rewrite H1 in H0; rewrite H0 in H; simpl (Z_of_nat 1) in H.
simpl; case H; auto with zarith.
apply zeroness_unh_aux2; trivial.
rewrite H1 in H0; rewrite H0 in H; simpl (Z_of_nat 1) in H.
simpl; case H; auto with zarith.
simpl.
rewrite H0 in H; rewrite H1 in H.
repeat rewrite inj_S in H.
unfold Ah, ubar, discretization.
repeat rewrite (zeroness_usol c u0 u1 f usol A B); trivial.
rewrite (fh_is_f dX).
replace (Xjk dX j (S n0)) with ((fst ((Xjk dX j (S n0)))), snd ((Xjk dX j (S n0)))) by reflexivity.
rewrite f_zero.
unfold Rdiv; ring.
apply zeroness_unh_aux2; trivial.
assert (Z_of_nat (S n0)*  (up (c * snd dX / fst dX))
  <= Zsucc (Zsucc (Z_of_nat n0)) *  (up (c * snd dX / fst dX)))%Z.
rewrite inj_S; apply Zmult_le_compat_r; omega.
omega.
apply zeroness_unh_aux2; trivial.
assert ((Z_of_nat (n0)+1)*  (up (c * snd dX / fst dX))
  < Zsucc (Zsucc (Z_of_nat n0)) *  (up (c * snd dX / fst dX)))%Z.
apply Zmult_lt_compat_r; omega.
rewrite inj_plus; simpl; omega.
apply zeroness_unh_aux2; trivial.
assert ((Z_of_nat (n0)+1)*  (up (c * snd dX / fst dX))
  < Zsucc (Zsucc (Z_of_nat n0)) *  (up (c * snd dX / fst dX)))%Z.
apply Zmult_lt_compat_r; omega.
rewrite inj_plus; simpl; omega.
apply zeroness_unh_aux2; trivial.
assert ((Z_of_nat (n0))*  (up (c * snd dX / fst dX))
  < Zsucc (Zsucc (Z_of_nat n0)) *  (up (c * snd dX / fst dX)))%Z.
apply Zmult_lt_compat_r; omega.
omega.
apply zeroness_unh_aux2; trivial.
assert ((Z_of_nat (n0)+1)*  (up (c * snd dX / fst dX))
  < Zsucc (Zsucc (Z_of_nat n0)) *  (up (c * snd dX / fst dX)))%Z.
apply Zmult_lt_compat_r; omega.
rewrite inj_plus; simpl; omega.
apply zeroness_unh_aux2; trivial.
assert ((Z_of_nat (n0)+2)*  (up (c * snd dX / fst dX))
  <= Zsucc (Zsucc (Z_of_nat n0)) *  (up (c * snd dX / fst dX)))%Z.
apply Zmult_le_compat_r; omega.
rewrite inj_plus; simpl; omega.
Qed.


Lemma FS_truncation_error: forall dX k, (0 < fst dX) -> (0 < snd dX) ->
 FS (fun j => truncation_error c usol u0h u1h fh dX j k).
intros dX k L M.
exists (Int_part (A/fst dX) - Z_of_nat k*(up(c*snd dX/ fst dX)))%Z; 
  exists  (Int_part (B/fst dX) + Z_of_nat k*(up (c*snd dX/ fst dX)))%Z.
apply FS_truncation_error_aux; trivial.
Qed.

Hint Resolve FS_truncation_error: FS.

(** Initialization is ok *)

Lemma LL1h_is_second_order: 
   is_sufficiently_regular_n 4%nat usol ->
  Oups (fun x dX => LL1h c usol (x,0) dX) 
           (fun dX => fst dX ^2 + snd dX^2).
intros.
unfold LL1h.
apply OuP_eq_trans_l with (fun x => add2_fun 
  (fun dX => (usol (fst (x, 0), snd (x, 0) + snd dX) - usol (x, 0)) / snd dX -
    partial_derive_secondvar usol (x, 0)
    - c*c*snd dX*/2 * partial_derive_firstvar_n 2 usol (x,0))
  (scalar_multiplyA_fun _ (c*c/2)  
     (fun dX  => snd dX * (partial_derive_firstvar_n 2 usol (x,0) 
       -(usol (fst (x, 0) + fst dX, snd (x, 0)) - 2 * usol (x, 0) +
         usol (fst (x, 0) - fst dX, snd (x, 0))) / (fst dX * fst dX))))).
intros; unfold add2_fun, scalar_multiplyA_fun.
unfold Rdiv; ring.
apply OuP_is_EV_plus.
(* DL of order 2 of usol in first variable *)
elim usol_is_sol; unfold L; intros T1 _.
apply OuP_eq_trans_l with
 (fun (a : R) =>
   (add2_fun (fun dX => (usol (fst (a, 0), snd (a, 0) + snd dX) - usol (a, 0)) / snd dX -
     partial_derive_secondvar usol (a, 0) -
     snd dX * / 2 * partial_derive_secondvar_n 2 usol (a, 0)) 
              (fun dX => snd dX * /2 * f (a,0)))).
intros; unfold add2_fun; rewrite <- T1; ring.
clear T1.
apply OuP_is_EV_plus.
apply OuP_multiply2.
intros dX (K1,K2); auto with real.
apply OuP_X_to_x.
apply OuP_eq_trans_l with 
 (fun X : R * R =>
          subtract2_fun (fun dX : R * R => usol (add2 (fst X,0)  (0,snd dX)))
            (fun dX : R * R => DL_pol (3 - 1) usol (fst X,0) (0,snd dX))).
intros a dX (K1,K2).
unfold subtract2_fun, DL_pol, differential, C, partial_derive.
simpl.
replace (add2 (fst a, 0) (0, snd dX)) with (fst a, 0 + snd dX).
field; auto with real.
unfold add2; simpl; now rewrite Rplus_0_r.
apply OuP_X_to_appr with 
  (appr:= fun X dX => (fst X,0))
  (f:= fun X => 
     subtract2_fun (fun dX : R * R => usol (add2 X (0, snd dX)))
     (fun dX : R * R => DL_pol (3 - 1) usol X (0, snd dX))).
apply OuP_inclusion with (fun dX => norm_l2 (0, snd dX) ^ 3).
intros X H1 (H2,H3).
rewrite norm_l2_proj2.
repeat rewrite Rabs_right; try apply Rle_ge; auto with real.
rewrite Rmult_plus_distr_l.
apply Rle_trans with (0+snd X ^ 3); auto with real.
apply Rplus_le_compat.
apply Rmult_le_pos; auto with real.
right; ring.
apply Rmult_le_pos; auto with real.
apply Rle_trans with (0+0); auto with real.
apply Rplus_le_compat; auto with real.
apply OuP_implies_OuP with (P1:=(fun _ => True)); trivial.
apply OuP_snddX_to_dX with 
  (h:= fun X => subtract2_fun (fun dX : R * R => usol (add2 X dX))
     (fun dX : R * R => DL_pol (3 - 1) usol X dX))
  (g:=fun dX => norm_l2 dX ^ 3).
trivial.
apply H ; auto.
apply OuP_eq_trans_l with (fun _ _ => 0).
intros; rewrite f_is_zero_at_beginning; ring.
apply OuP_is_EV_zero.
(* discrete partial second order derivative is nearly the continuous one *)
apply OuP_is_EV_mult.
apply OuP_multiply1.
intros dX (H1,H2); auto with real.
apply OuP_multiply1.
intros dX (H1,H2); auto with real.
apply OuP_divide2.
intros dX (H1,H2); auto with real.
apply OuP_eq_trans_l with 
 (fun (X : R) (dX : R * R) =>
      (fst dX^2 * partial_derive_firstvar_n 2 usol (X, 0) -
       (usol (X + fst dX,  0) - 2 * usol (X, 0) +
        usol (X - fst dX,  0)))).
intros a dX (H1,H2); simpl; field; split; auto with real.
apply OuP_X_to_x.
apply OuP_eq_trans_l with
  (scalar_multiplyAA_fun _ _ (-1)
  (fun X dX  =>
          (usol (add2 (fst X,0)  (fst dX, 0)) - 2 * usol (fst X,0) + usol (subtract2 (fst X,0) (fst dX, 0)) -
           fst dX ^ 2 * partial_derive_firstvar_n 2 usol (fst X,0))%R)).
intros.
unfold scalar_multiplyAA_fun, add2, subtract2; simpl.
rewrite Rplus_0_r; rewrite Rminus_0_r; ring.
apply OuP_multiplyAA.
apply OuP_inclusion with  (fun dX : R * R => fst dX ^ 4).
intros X H1 (H2,H3).
replace (/ snd X * (fst X * (fst X * (fst X ^ 2 + snd X ^ 2)))) with
   ( fst X ^ 4 */ snd X + fst X ^ 2 * snd X).
repeat rewrite Rabs_right.
apply Rle_trans with (fst X ^ 4 * /1 +0)%R;[right; field|idtac].
apply Rplus_le_compat.
apply Rmult_le_compat_l.
auto with real.
apply Rle_Rinv; auto with real.
apply Rle_trans with (2:=H1).
apply Rle_trans with (2:=Rabs_snd_le_norm_l2 X).
rewrite Rabs_right; try apply Rle_ge; auto with real.
apply Rmult_le_pos; auto with real.
apply Rle_ge; apply Rle_trans with (0+0); auto with real.
apply Rplus_le_compat; apply Rmult_le_pos; auto with real.
apply Rle_ge; auto with real.
field; auto with real.
apply OuP_X_to_appr with (appr:= fun X dX => (fst X,0))
   (f:= fun X dX => (usol (add2 X  (fst dX, 0)) - 2 * usol X +
    usol (subtract2 X  (fst dX, 0)) -
    fst dX ^ 2 * partial_derive_firstvar_n 2 usol X)%R). 
apply OuP_implies_OuP with (P1:=(fun _ => True)); trivial.
now apply DL_discretization_eps_dx.
Qed.   

(** Main lemma: the truncation error is O(dx^2+dt^2) *)
(** Proof by Taylor series and *many* computations *)

Lemma pointwise_consistency:
 is_sufficiently_regular_n 4 usol ->
 Oups (fun (xt:R*R) dxt => let (x,t):= xt in let (dx,dt) := dxt in
                         truncation_error c usol u0h u1h fh dxt (jx x dx)  (kt t dt)) 
      (fun dX =>  (fst dX)^2 + (snd dX)^2).
intros H.
apply OuP_eq_trans_l with
  (fun X dX => truncation_error c usol u0h u1h fh dX (jx (fst X) (fst dX)) (kt (snd X) (snd dX))).
intros a dX; destruct a; destruct dX; simpl; auto.
unfold truncation_error, scheme, Lh.
apply OuP_eq_trans_l with
  (fun X dX => match kt (snd X) (snd dX) with
   | 0%nat =>
       ubar usol dX (jx (fst X) (fst dX)) 0 - u0h (jx (fst X) (fst dX))
   | S n => match (pred (S n)) with 
     | 0%nat =>
       (ubar usol dX (jx (fst X) (fst dX)) 1 -
        ubar usol dX (jx (fst X) (fst dX)) 0) / snd dX +
       snd dX / 2 *
       Ah c (fun j' : Z => ubar usol dX j' 0) (fst dX) (jx (fst X) (fst dX)) -
       u1h (jx (fst X) (fst dX))
      | (S n0) =>
           (ubar usol dX (jx (fst X) (fst dX)) (S (S n0)) -
            2 * ubar usol dX (jx (fst X) (fst dX)) (S n0) +
            ubar usol dX (jx (fst X) (fst dX)) n0) / (snd dX * snd dX) +
           Ah c (fun j' : Z => ubar usol dX j' (S n0)) (fst dX)
             (jx (fst X) (fst dX))
        - fh (jx (fst X) (fst dX)) (S n0)
      end
   end).
intros a dX Y.
case_eq (kt (snd a) (snd dX)); intros.
reflexivity.
replace (pred (S n)) with n by omega.
case_eq n; intros.
reflexivity.
replace  (S (S n0)) with (n0+2)%nat by omega.
replace  (S n0) with (n0+1)%nat by omega.
reflexivity.
apply OuP_match with (f1:= fun n X dX => match pred n return R with
       | 0%nat =>
           (ubar usol dX (jx (fst X) (fst dX)) 1 -
            ubar usol dX (jx (fst X) (fst dX)) 0) / snd dX +
           snd dX / 2 *
           Ah c (fun j' : Z => ubar usol dX j' 0) (fst dX)
             (jx (fst X) (fst dX)) - u1h (jx (fst X) (fst dX))
       | S n0 =>
           (ubar usol dX (jx (fst X) (fst dX)) (S (S n0)) -
            2 * ubar usol dX (jx (fst X) (fst dX)) (S n0) +
            ubar usol dX (jx (fst X) (fst dX)) n0) / (snd dX * snd dX) +
           Ah c (fun j' : Z => ubar usol dX j' (S n0)) (fst dX)
             (jx (fst X) (fst dX)) - fh (jx (fst X) (fst dX)) (S n0)
       end).
(* n = 0 *)
apply OuP_eq_trans_l with (fun _ _ => 0).
intros.
rewrite u0h_is_u0 with dX (jx (fst a) (fst dX)).
destruct usol_is_sol as (_,(_,H1)).
unfold L0 in H1; rewrite <- H1.
unfold ubar, discretization, Xjk; simpl.
rewrite Rmult_0_l; ring.
apply OuP_is_EV_zero.
apply OuP_match with (f1 := fun n X dX =>
    (ubar usol dX (jx (fst X) (fst dX)) (S n) -
        2 * ubar usol dX (jx (fst X) (fst dX)) n +
        ubar usol dX (jx (fst X) (fst dX)) (pred n)) / (snd dX * snd dX) +
       Ah c (fun j' : Z => ubar usol dX j' n) (fst dX)
         (jx (fst X) (fst dX)) - fh (jx (fst X) (fst dX)) n).
(* n = 1 *)
apply OuP_eq_trans_l with (fun X dX => LL1h c usol (Xxt (fst X,0) dX) dX).
intros.
rewrite u1h_is_u1 with dX (jx (fst a) (fst dX)).
destruct usol_is_sol as (_,(H1,_)).
unfold L1 in H1; rewrite <- H1.
unfold LL1h, Ah, Xxt.
unfold ubar, discretization, Xjk.
simpl.
replace (INR (kt 0 (snd dX))) with 0%R.
rewrite Rmult_0_l.
rewrite Rplus_0_l.
rewrite Rmult_1_l.
replace (IZR (jx (fst a) (fst dX) + 1) * fst dX) with 
(IZR (jx (fst a) (fst dX)) * fst dX + fst dX).
replace (IZR (jx (fst a) (fst dX) - 1) * fst dX) with
  (IZR (jx (fst a) (fst dX)) * fst dX - fst dX).
field.
elim H0; split; auto with real.
unfold Zminus; rewrite plus_IZR; simpl; ring.
rewrite plus_IZR; simpl; ring.
unfold kt.
replace (0/snd dX) with (IZR 0).
rewrite Int_part_IZR.
now simpl.
unfold Rdiv; simpl; ring.
apply OuP_X_to_appr_fst.
intros; unfold Xxt, kt; simpl.
replace (0 / snd dX) with (IZR 0).
rewrite Int_part_IZR.
simpl.
now rewrite Rmult_0_l.
unfold Rdiv; simpl; ring.
now apply LL1h_is_second_order.
(* autres n *)
apply OuP_eq_trans_l with
(fun X dX : R * R =>
   ((ubar usol dX (jx (fst X) (fst dX))
       (S (S (pred (pred (kt (snd X) (snd dX)))))) -
     2 *
     ubar usol dX (jx (fst X) (fst dX))
       (S (pred (pred (kt (snd X) (snd dX))))) +
     ubar usol dX (jx (fst X) (fst dX))
       (pred (pred (kt (snd X) (snd dX))))) /
    (snd dX * snd dX) +
    Ah c
      (fun j' : Z =>
       ubar usol dX j' (S (pred (pred (kt (snd X) (snd dX))))))
      (fst dX) (jx (fst X) (fst dX)) -
    fh (jx (fst X) (fst dX))
      (S (pred (pred (kt (snd X) (snd dX)))))))%R.
intros.
now repeat rewrite <- pred_Sn.
Notation blop :=  (fun X dX => (S (pred (pred (kt (snd X) (snd dX)))))).
Notation Xxtd :=  (fun X dX => (Xjk dX (jx (fst X) (fst dX)) (S (pred (pred (kt (snd X) (snd dX))))))).
apply OuP_eq_trans_l with (fun X: R * R => add2_fun 
   (fun dX => (((ubar usol dX (jx (fst X) (fst dX)) ((blop X dX) + 1)%nat -
      2 * ubar usol dX (jx (fst X) (fst dX)) (blop X dX) +
      ubar usol dX (jx (fst X) (fst dX)) (blop X dX - 1)) / (snd dX * snd dX))
         -  partial_derive_secondvar_n 2 usol (Xxtd X dX)))
   (fun dX => ((Ah c (fun j : Z => ubar usol dX j (blop X dX)) (fst dX) (jx (fst X) (fst dX)))
          + c ^ 2 * partial_derive_firstvar_n 2 usol (Xxtd X dX)))).
intros; destruct a; destruct dX; unfold add2_fun.
replace (partial_derive_secondvar_n 2 usol (Xxtd (r, r0) (r1, r2))) with
  (c ^ 2 * partial_derive_firstvar_n 2 usol (Xxtd (r, r0) (r1, r2))
    + f (Xxtd (r, r0) (r1, r2))).
simpl; rewrite (fh_is_f (r1,r2)).
replace (S (pred (pred (kt r0 r2)) + 1)) with (S (S (pred (pred (kt r0 r2))))) by omega.
replace (pred (pred (kt r0 r2)) - 0)%nat with (pred (pred (kt r0 r2))) by omega.
ring.
destruct usol_is_sol as (T1,(T2,T3)); unfold L in T1.
rewrite <- T1; ring.
apply OuP_is_EV_plus.
(* d\u00e9riv\u00e9e en temps *)
apply OuP_inclusion with (fun dX : R * R => snd dX ^ 2).
intros.
unfold pow; repeat rewrite Rmult_1_r.
fold (Rsqr (snd X)); fold (Rsqr (fst X)).
repeat rewrite Rabs_right; try apply Rle_ge; auto with real.
apply Rle_trans with (0 + Rsqr (snd X) ); auto with real.
apply Rle_trans with (0 + 0); auto with real.
apply OuP_multiply2.
intros dX (H1,H2); auto with real.
apply OuP_multiply2.
intros dX (H1,H2); auto with real.
apply OuP_eq_trans_l with 
 (fun X dX : R * R =>
     ((ubar usol dX (jx (fst X) (fst dX)) ((blop X dX) + 1) -
       2 * ubar usol dX (jx (fst X) (fst dX)) (blop X dX) +
       ubar usol dX (jx (fst X) (fst dX)) (blop X dX - 1)) 
     - snd dX^2 * partial_derive_secondvar_n 2 
        usol (Xxtd X dX))).
intros a dX (H1,H2); field; auto with real.
apply OuP_eq_trans_r with (fun dX : R * R => snd dX ^ 4).
intros; ring.
unfold ubar, discretization, Xxt.
apply OuP_eq_trans_l with 
  (fun X dX : R * R =>
   (usol (add2 (Xjk dX (jx (fst X) (fst dX)) (S (pred (pred (kt (snd X) (snd dX)))))) (0, snd dX)) -
    2 * usol (Xjk dX (jx (fst X) (fst dX)) (S (pred (pred (kt (snd X) (snd dX)))))) +
    usol (subtract2 (Xjk dX (jx (fst X) (fst dX)) (S (pred (pred (kt (snd X) (snd dX)))))) (0, snd dX)) -
    snd dX ^ 2 * partial_derive_secondvar_n 2 usol
      (Xjk dX (jx (fst X) (fst dX)) (S (pred (pred (kt (snd X) (snd dX)))))))).
intros.
replace (add2 (Xjk dX (jx (fst a) (fst dX)) (blop a dX)) (0, snd dX)) with
   (Xjk dX (jx (fst a) (fst dX)) (blop a dX + 1)).
replace (subtract2 (Xjk dX (jx (fst a) (fst dX)) (blop a dX)) (0, snd dX)) with
   (Xjk dX (jx (fst a) (fst dX)) (blop a dX - 1)).
ring.
unfold subtract2, Xjk.
replace (S (pred (pred (kt (snd a) (snd dX)))) - 1)%nat with
   (pred (pred (kt (snd a) (snd dX)))) by omega.
simpl.
replace (IZR (jx (fst a) (fst dX)) * fst dX - 0) with (IZR (jx (fst a) (fst dX)) * fst dX) by ring.
case (pred (pred (kt (snd a) (snd dX)))).
replace (INR 0 * snd dX) with (1 * snd dX - snd dX) by (simpl; ring); auto.
intros.
replace (INR (S n) * snd dX) with ((INR (S n) + 1) * snd dX - snd dX) by ring; auto.
unfold add2, Xjk.
replace (snd
  (IZR (jx (fst a) (fst dX)) * fst dX,
  INR (S (pred (pred (kt (snd a) (snd dX))))) * snd dX) + snd (0, snd dX)) with
  (INR (S (pred (pred (kt (snd a) (snd dX)))) + 1) * snd dX).
simpl.
replace (IZR (jx (fst a) (fst dX)) * fst dX + 0) with (IZR (jx (fst a) (fst dX)) * fst dX) by ring; auto.
rewrite plus_INR; simpl; ring.
apply OuP_implies_OuP with (P1:=fun _ => True).
intros; auto.
apply OuP_X_to_appr with (appr:= fun X dX => (Xjk dX (jx (fst X) (fst dX)) (S (pred (pred (kt (snd X) (snd dX)))))))
  (f := fun X dX => usol (add2 X (0, snd dX)) -2 *usol X + usol (subtract2 X (0, snd dX))
     -  snd dX ^ 2 * partial_derive_secondvar_n 2 usol X).
now apply DL_discretization_eps_dt with (f:=usol).
(* d\u00e9riv\u00e9e en espace *)
apply OuP_inclusion with (fun dX : R * R => fst dX ^ 2).
intros.
unfold pow; repeat rewrite Rmult_1_r.
fold (Rsqr (snd X)); fold (Rsqr (fst X)).
repeat rewrite Rabs_right; try apply Rle_ge; auto with real.
apply Rle_trans with (Rsqr (fst X)+0); auto with real.
apply Rle_trans with (0 + 0); auto with real.
unfold Ah.
apply OuP_eq_trans_l with (scalar_multiplyAA_fun _ _ (-c*c) 
   (fun a dX => (ubar usol dX (jx (fst a) (fst dX) + 1) (blop a dX) -
    2 * ubar usol dX (jx (fst a) (fst dX)) (blop a dX) +
    ubar usol dX (jx (fst a) (fst dX) - 1) (blop a dX)) /
   (fst dX * fst dX) 
  - partial_derive_firstvar_n 2 usol (Xxtd a dX))).
intros; unfold scalar_multiplyAA_fun; simpl.
unfold Rdiv; ring.
apply OuP_is_EV_mult.
apply OuP_multiply1.
intros dX (H1,H2); auto with real.
apply OuP_multiply1.
intros dX (H1,H2); auto with real.
apply OuP_eq_trans_l with 
 (fun X dX : R * R =>
     ((ubar usol dX (jx (fst X) (fst dX) + 1) (blop X dX) -
       2 * ubar usol dX (jx (fst X) (fst dX)) (blop X dX) +
       ubar usol dX (jx (fst X) (fst dX) - 1) (blop X dX)) 
     - fst dX^2 * partial_derive_firstvar_n 2 
        usol (Xxtd X dX))).
intros a dX (H1,H2); field; auto with real.
apply OuP_eq_trans_r with (fun dX : R * R => fst dX ^ 4).
intros; ring.
unfold ubar, discretization, Xxt.
apply OuP_eq_trans_l with 
  (fun X dX : R * R =>
   (usol (add2 (Xjk dX (jx (fst X) (fst dX)) (blop X dX)) (fst dX,0)) -
    2 * usol (Xjk dX (jx (fst X) (fst dX)) (blop X dX)) +
    usol (subtract2 (Xjk dX (jx (fst X) (fst dX)) (blop X dX)) (fst dX,0)) -
    fst dX ^ 2 * partial_derive_firstvar_n 2 usol
      (Xjk dX (jx (fst X) (fst dX)) (blop X dX)))).
intros.
replace (add2 (Xjk dX (jx (fst a) (fst dX)) (blop a dX)) (fst dX, 0)) with
   (Xjk dX (jx (fst a) (fst dX) + 1) (blop a dX)).
replace (subtract2 (Xjk dX (jx (fst a) (fst dX)) (blop a dX)) (fst dX, 0)) with
   (Xjk dX (jx (fst a) (fst dX) - 1) (blop a dX)).
ring.
unfold add2, subtract2, Xjk; simpl.
unfold Zminus; rewrite plus_IZR; simpl.
replace ((IZR (jx (fst a) (fst dX)) + -1) * fst dX) with 
  (IZR (jx (fst a) (fst dX)) * fst dX + - fst dX) by ring; auto with real.
unfold add2, Xjk; simpl.
rewrite plus_IZR; simpl.
replace ((IZR (jx (fst a) (fst dX)) + 1) * fst dX) with 
  (IZR (jx (fst a) (fst dX)) * fst dX + fst dX) by ring; auto with real.
apply OuP_implies_OuP with (P1:=fun _ => True).
intros; auto.
apply OuP_X_to_appr with (appr:= fun X dX => (Xjk dX (jx (fst X) (fst dX)) (blop X dX)))
  (f := fun X dX => usol (add2 X (fst dX, 0)) -2 *usol X + usol (subtract2 X (fst dX, 0))
     -  fst dX ^ 2 * partial_derive_firstvar_n 2 usol X).
now apply DL_discretization_eps_dx with (f:=usol).
Qed.

(** From one point to an infinite squared norm *)

Lemma Oup_pointwise_to_squared_norm_dx : forall vh, forall g,
  (forall (dX : R * R),
    let (dx, dt) := dX in
    (0 < dx) ->
    (0 < dt) ->
    (forall k j,
      (j < Int_part (A / dx) - Z_of_nat k * (up (c * dt / dx)) \/
        Int_part (B / dx) + Z_of_nat k * (up (c * dt / dx)) < j)%Z -> 
    vh dX j k = 0)) ->
  Oups
    (fun (X : R * R) dX =>
      let (x, t) := X in
      let (dx, dt) := dX in
      vh dX (jx x dx) (kt t dt))
    g ->
  OupsCFL_l
    (fun (t : {t|Under T t}) dX =>
      infinite_squared_norm_dx (fst dX) (fun j => vh dX j (kt (projT1 t) (snd dX))))
    (fun dX => Rsqr (g dX)).
Proof.
intros.
destruct H0 as (alp, (C, H1)); intuition.
exists alp; exists (C * C * (IZR (Int_part B - Int_part A) + 2 * c * T + 2 * c * T / eta + 2 * alp + 1)).
split; trivial.
split.
repeat apply Rmult_lt_0_compat; trivial.
apply Rplus_le_lt_0_compat; auto with real.
repeat apply Rplus_le_le_0_compat; try apply Rmult_le_pos; auto with real.
replace 0%R with (IZR 0) by reflexivity.
apply IZR_le.
assert (Int_part A <=  Int_part B)%Z; auto with zarith.
apply Int_part_monotone; auto.
apply Rmult_le_pos; auto with real.
left; apply T_pos.
repeat apply Rmult_le_pos; auto with real.
left; apply T_pos.

intros t (dx, dt) H4; simpl; intros (Y1, (Y2, Y3)).
rewrite Rabs_right.
apply Rle_trans with
  (dx * IZR ((Int_part (B / dx) + Z_of_nat (kt (projT1 t) dt) * (up (c * dt / dx))) -
  (Int_part (A / dx) - Z_of_nat (kt (projT1 t) dt) * (up (c * dt / dx))) + 1) *
  (Rsqr (C * (g (dx, dt))))).
apply infinite_squared_norm_dx_majoration; auto with real.

assert (0 <= Z_of_nat (kt (projT1 t) dt) * (up (c * dt / dx)))%Z.
apply Zmult_le_0_compat; auto with zarith.
assert (0 < up (c * dt / dx))%Z; auto with zarith.
apply up_gt_zero.
repeat apply Rmult_lt_0_compat; auto with real.

apply Zle_trans with (Int_part (A / dx)); auto with zarith.
apply Zle_trans with (Int_part (B / dx)); auto with zarith.
apply Int_part_monotone.
unfold Rdiv; apply Rmult_le_compat_r; auto with real.
exact (H (dx, dt) Y1 Y2 (kt (projT1 t) dt)).
intros.
rewrite Rsqr_abs; rewrite (Rsqr_abs (C * g (dx, dt))).
apply Rsqr_incr_1; try apply Rabs_pos.
replace j with (jx (IZR j * dx) dx).
rewrite Rabs_mult.
rewrite (Rabs_right C).
apply H3 with (X := ((IZR j * dx), projT1 t)) (dX := (dx, dt)); auto.
apply Rle_ge; auto with real.
unfold jx.
replace (IZR j * dx / dx) with (IZR j).
apply Int_part_IZR.
field; auto with real.

apply Rle_trans with 
  ((IZR (Int_part B - Int_part A) + 2 * c * T + 2 * c * T / eta + 2 * alp + 1) *
  (C * C * (Rabs (Rsqr (g (dx, dt)))))).
2: right; ring.
rewrite Rsqr_mult.
unfold Rsqr at 1.
rewrite Rabs_right; auto with real.
apply Rmult_le_compat_r.
apply Rmult_le_pos; auto with real.
apply Rmult_le_pos; auto with real.
unfold Zminus; repeat rewrite plus_IZR; repeat rewrite Ropp_Ropp_IZR.
rewrite plus_IZR; rewrite Ropp_Ropp_IZR.
rewrite mult_IZR.
rewrite <- INR_IZR_INZ.
apply Rle_trans with
  ((dx * (IZR (Int_part (B / dx))) - 1) + 
  (dx * (- 1 - IZR (Int_part (A / dx))) +
  (2 * dx * INR (kt (projT1 t) dt)  * IZR (up (c * dt / dx))+ 
  2 * dx + 1))).
right; simpl; ring.
apply Rle_trans with (IZR (Int_part B) +(- IZR (Int_part A) + (2 * c * T + 2 * c * T / eta + 2 * alp + 1))).
2: right; ring.
apply Rplus_le_compat.
apply Rle_trans with (dx * (B / dx) - 1).
apply Rplus_le_compat_r.
apply Rmult_le_compat_l; auto with real.
eapply base_Int_part.
replace (dx * (B / dx)) with B by (field; auto with real).
apply Rplus_le_reg_l with (- B); ring_simplify.
rewrite Rplus_comm; left.
eapply base_Int_part.
apply Rplus_le_compat.
apply Ropp_le_cancel; rewrite Ropp_involutive.
apply Rle_trans with (1 := proj1 (base_Int_part A)).
apply Rmult_le_reg_l with (/ dx); auto with real.
apply Rle_trans with (IZR (Int_part (A / dx)) + 1).
2: right; field; auto with real.
apply Rplus_le_reg_l with (- 1 - A * / dx); ring_simplify.
rewrite Ropp_mult_distr_l_reverse.
rewrite Rplus_comm; left.
apply (proj2 (base_Int_part (A * / dx))).
apply Rplus_le_compat_r.
apply Rplus_le_compat.
rewrite (Rmult_assoc 2 c T).
unfold Rdiv at 2.
rewrite (Rmult_assoc 2 (c * T) (/ eta)).
rewrite <- Rmult_plus_distr_l.
repeat rewrite Rmult_assoc. 
apply Rmult_le_compat_l; auto with real.
apply Rle_trans with (dx * ((T / dt) * (c * dt / dx + 1))).
apply Rmult_le_compat_l; auto with real.
apply Rmult_le_compat; auto with real.
assert (0 < up (c * dt / dx))%Z.
apply up_gt_zero.
repeat apply Rmult_lt_0_compat; auto with real.
replace 0 with (IZR 0); auto with real zarith.
apply IZR_le; auto with zarith.
unfold kt.
rewrite INR_IZR_INZ; rewrite inj_Zabs_nat.
rewrite Zabs_eq.
apply Rle_trans with ((projT1 t / dt)).
eapply base_Int_part.
unfold Rdiv; apply Rmult_le_compat_r; auto with real.
destruct t; simpl; eapply u.
apply Int_part_pos.
unfold Rdiv; apply Rmult_le_pos; auto with real.
destruct t; simpl; eapply u.
apply Rplus_le_reg_l with (- (c * dt / dx)); ring_simplify.
rewrite Rplus_comm; eapply archimed.
rewrite <- Rmult_assoc; rewrite Rmult_plus_distr_l.
apply Rplus_le_compat.
right; field; auto with real.
unfold CFL_condition_l, CFL in Y3; simpl in Y3.
replace (dx * (T / dt) * 1) with (T * (dx / dt)) by (unfold Rdiv; ring).
replace (c * (T * / eta)) with (T * (c * / eta)) by ring.
apply Rmult_le_compat_l; auto with real.
left; apply T_pos.
apply Rmult_le_reg_l with (eta * dt * / dx).
repeat apply Rmult_lt_0_compat; auto with real.
apply Rle_trans with eta.
right; field; auto with real.
apply Rle_trans with (1 := Y3).
right; field; auto with real.
apply Rmult_le_compat_l; auto with real.
rewrite <- (Rabs_right dx).
replace dx with (fst (dx, dt)); auto.
apply Rle_trans with (1 := Rabs_fst_le_norm_l2 (dx, dt)); auto with real.
apply Rle_ge; auto with real.
apply Rle_ge; apply infinite_squared_norm_dx_pos; auto with real.

unfold FS.
exists (Int_part (A / dx) - Z_of_nat (kt (projT1 t) dt) * up (c * dt / dx))%Z;
  exists (Int_part (B / dx) + Z_of_nat (kt (projT1 t) dt) * up (c * dt / dx))%Z.
exact (H (dx, dt) Y1 Y2 (kt (projT1 t) dt)).
Qed.

Lemma Oup_pointwise_to_squared_norm_dx_truncation_error: forall g,
  Oups
    (fun (X : R * R) dX =>
      let (x, t) := X in
      let (dx, dt) := dX in
      truncation_error c usol u0h u1h fh dX (jx x dx) (kt t dt)) 
    g ->
  OupsCFL_l
    (fun (t : {t|Under T t}) dX =>
      (*let (dx, dt) := dX in*)
      infinite_squared_norm_dx (fst dX)
        (fun j => truncation_error c usol u0h u1h fh dX j (kt (projT1 t) (snd dX))))
    (fun dX => Rsqr (g dX)).
Proof.
intros.
apply (Oup_pointwise_to_squared_norm_dx (truncation_error c usol u0h u1h fh)).
2: exact H.
destruct dX; intros.
apply FS_truncation_error_aux; trivial.
Qed.

Theorem consistency: 
  is_sufficiently_regular_n 4%nat usol -> 
  scheme_is_consistent_of_order c usol u0h u1h fh eta 2%nat 2%nat.
Proof.
intros; unfold scheme_is_consistent_of_order; intros.
apply
  (Oup_pointwise_to_squared_norm_dx_truncation_error
    (fun dX : R * R => (fst dX) ^ 2 + (snd dX) ^ 2)); auto.
apply pointwise_consistency; auto.
Qed.

End Consistency. 
