(** Stability of the numerical scheme *)

Require Import Fourier.
Require Export one_d_wave_equation.

Open Local Scope nat_scope.
Open Local Scope R_scope.

Section Stab.

(** Here, we take one particular dX for the whole file.
    It will go towards zero only in convergence.v. *)

Variable c:R.
Variable u0h u1h: Z-> R.
Variable fh: Z -> nat -> R.
Variable unh : R * R -> Z -> nat -> R.
Variable usol : R * R -> R.
Variable eps:R.
Variable dX:R*R.
Variable l j1 j2:Z.

Hypothesis unh_is_discrete_sol: is_discrete_solution c u0h u1h fh unh dX.
Hypothesis eps_lt: 0 < eps < 1.
Hypothesis CFL_eps: CFL_condition_r c eps dX.
Hypothesis c_pos: 0 < c.
Hypothesis l_pos: (0 < l)%Z.

Hypothesis u0h_zero: forall j, (j < j1)%Z  \/ ( j2 < j)%Z 
   -> u0h j = 0.
Hypothesis u1h_zero: forall j, (j < j1)%Z  \/ ( j2 < j)%Z 
   -> u1h j = 0.
Hypothesis fh_zero: forall j k, (j < j1-Z_of_nat k*l)%Z  \/ ( j2+Z_of_nat k*l < j)%Z 
   -> fh j k= 0.


Lemma  FS_fh: forall k, FS (fun j => fh j k).
intros; exists (j1-Z_of_nat k*l)%Z; exists (j2+Z_of_nat k*l)%Z; auto.
Qed.


Lemma FS_unh: forall (n:nat), snd dX <> 0 ->
     FS (fun j => unh dX j n).
intros; exists (j1 - Z_of_nat n*l)%Z; exists (j2 + Z_of_nat n*l)%Z.
apply zeroness_unh_h with c u0h u1h fh; auto.
Qed.

Hint Resolve FS_fh FS_unh  (FS_Ah c) (FS_Ah_fun c): FS.

(** Evolution of the energy when time increases *)

Lemma energy_increases: forall  n:nat,
   snd dX <> 0 -> (0<n)%nat->
   (energy c (unh dX) dX n) - (energy c (unh dX) dX (n-1)%nat) =
  1/2*infinite_dot_product_dx (fst dX) (fun j => fh j n) 
            (fun j => unh dX j (n+1)%nat - unh dX j (n-1)%nat).
intros n H1 H2.
rewrite infinite_dot_product_dx_eq_r with 
 (v:=subtract_n (fun j => unh dX j (n+1)%nat)  (fun j => unh dX j (n-1)%nat)); FS.
unfold energy.
repeat rewrite <- Rmult_plus_distr_l in |- *.      
rewrite <- Rmult_minus_distr_l in |- *. 
apply Rmult_eq_compat_l.
unfold infinite_squared_norm_dx.
rewrite infinite_dot_product_dx_eq_l with (u:=(fun j : Z => fh j n)) 
  (v:=(add_n (scalar_multiply_n (1/(snd dX*snd dX)) 
             (subtract_n (add_n (fun j =>unh dX j (n+1)%nat) 
                            (fun j =>unh dX j (n-1)%nat))
                       (scalar_multiply_n 2 (fun j=> unh dX j n)))) 
               (fun j => (Ah c (fun j' => unh dX j' n) (fst dX) j)))); FS.
repeat rewrite  infinite_dot_product_dx_add_l; FS.
repeat rewrite infinite_dot_product_dx_scal_mult_l; FS.
repeat rewrite infinite_dot_product_dx_scal_mult_r; FS.
repeat rewrite  infinite_dot_product_dx_sub_l; FS.
repeat rewrite infinite_dot_product_dx_sub_r; FS.
repeat rewrite  infinite_dot_product_dx_add_l; FS.
replace (n - 1 +1)%nat with n;try omega.
repeat rewrite infinite_dot_product_dx_scal_mult_l; FS.
repeat rewrite  infinite_dot_product_dx_sub_l; FS.
repeat rewrite  infinite_dot_product_dx_sub_r; FS.
field_simplify;try assumption.
repeat rewrite (infinite_dot_product_dx_sym (fst dX) (fun j : Z => unh dX j n)); FS.
repeat rewrite (infinite_dot_product_dx_sym (fst dX) _ (fun j : Z => unh dX j (n-1))); FS.
rewrite (infinite_dot_product_dx_sym (fst dX) (fun j => unh dX j (n+1))  (fun j : Z => unh dX j (n-1))); FS.
field_simplify;try assumption.
unfold Rdiv;apply Rmult_eq_compat_r.
unfold Rminus;repeat rewrite Rplus_assoc.
repeat apply Rplus_eq_compat_l.
apply Rplus_eq_compat_r.
repeat apply f_equal.
rewrite infinite_dot_product_dx_eq_l with (v:= Ah c (fun j' : Z => unh dX j' (n - 1)) (fst dX))
(u:=fun j : Z => Ah c (fun j' : Z => unh dX j' (n - 1)) (fst dX) j); FS.
rewrite Ah_is_symmetrical with c (fst dX) (fun j'=> unh dX j' (n-1))
  (fun j : Z => unh dX j n); FS.
apply infinite_dot_product_dx_sym; FS.
intros i.
unfold add_n,scalar_multiply_n,subtract_n;simpl.
assert (exists m, n = S m).
exists (pred n); omega. 
destruct H as (m, Hm).	
repeat rewrite Hm.
replace (S m +1)%nat with (m+2)%nat by omega.
replace (S m -1)%nat with m by omega.
repeat replace (S m) with (m+1)%nat by omega.
generalize (unh_is_discrete_sol i (S (S m))); simpl.
intros T.
rewrite <- Rplus_0_l.
rewrite <- Ropp_0.
rewrite <- T.
field_simplify; try assumption.
replace (S m) with (m+1)%nat by omega.
field; assumption.
Save.

(** Energy is greater than... *)
Lemma minorationenergy:  forall n,
   snd dX <> 0 -> 0 < (fst dX)  ->
    1/2*(1-Rsqr (CFL c dX))*
  infinite_squared_norm_dx (fst dX) 
       (fun j=> 1/(snd dX)*(unh dX j (n+1)%nat - unh dX j n))
  <= energy c (unh dX) dX n.
intros.
unfold energy.
rewrite <- Rmult_plus_distr_l.
rewrite Rmult_assoc.
apply Rmult_le_compat_l.
fourier.
unfold Rminus at 1; rewrite Rmult_plus_distr_r.
apply Rplus_le_compat.
rewrite Rmult_1_l.
apply Rle_refl.
rewrite semi_infinite_dot_product_dx with (fst dX)
   (fun u j => Ah c u (fst dX) j)  (fun j' : Z => unh dX j' n)
   (fun j : Z => unh dX j (n + 1)); FS.
2: apply Ah_is_symmetrical.
unfold subtract_n, add_n.
apply Rle_trans with (-( /4 *
 infinite_dot_product_dx (fst dX)  
  (fun j : Z => Ah c (fun i : Z => unh dX i n - unh dX i (n + 1)) (fst dX) j)
   (fun i : Z => unh dX i n - unh dX i (n + 1) ))).
rewrite Ropp_mult_distr_l_reverse.
apply Ropp_le_contravar.
apply Rmult_le_reg_l with 4.
fourier.
rewrite <- Rmult_assoc.
rewrite Rinv_r.
2: discrR.
rewrite Rmult_1_l.
apply Rle_trans with
  (4 * c * c / ((fst dX) * (fst dX)) * infinite_squared_norm_dx (fst dX)
     (fun i : Z => unh dX i n - unh dX i (n + 1))).
apply Ah_cont; auto with FS.
rewrite infinite_squared_norm_dx_eq with
  (f2:=scalar_multiply_n  (-1 / snd dX)
      (fun i : Z => unh dX i n - unh dX i (n + 1)))
  (f1:= (fun j : Z => 1 / snd dX * 
     (unh dX j (n + 1) - unh dX j n))); FS.
rewrite infinite_squared_norm_dx_homogen; FS.
unfold CFL.
right; unfold Rsqr; field.
split; trivial; now apply Rgt_not_eq.
intros; unfold scalar_multiply_n.
field.
exact H.
rewrite  <- (Rplus_0_l  (-(/ 4 *
  infinite_dot_product_dx (fst dX) 
     (fun j : Z => Ah c (fun i : Z => unh dX i n - unh dX i (n + 1)) (fst dX) j)
     (fun i : Z => unh dX i n - unh dX i (n + 1) )))).
apply Rplus_le_compat_r.
apply Rmult_le_pos.
fourier.
apply Ah_pos; auto with FS.
Qed.



Lemma energy_pos: 
 forall n, (0 < snd dX) -> 0 < (fst dX)  ->
   0 <= energy c (unh dX) dX n.
intros.
assert (snd dX <> 0) by auto with real.
apply Rle_trans with ( 1/2*(1-Rsqr (CFL c dX))*
  infinite_squared_norm_dx (fst dX) 
       (fun j=> 1/(snd dX)*(unh dX j (n+1)%nat - unh dX j n))).
2: apply minorationenergy; trivial.
apply Rmult_le_pos.
repeat apply Rmult_le_pos.
auto with real.
fourier.
apply Rplus_le_reg_l with (Rsqr (CFL c dX)); ring_simplify.
replace 1 with (Rsqr 1) by (unfold Rsqr; ring).
apply Rsqr_incr_1.
apply Rle_trans with (1-eps).
apply CFL_eps.
apply Rplus_le_reg_l with (-1+eps); ring_simplify.
left; apply eps_lt.
unfold CFL; repeat apply Rmult_le_pos.
now left.
now left.
left; now apply Rinv_0_lt_compat.
auto with real.
apply infinite_squared_norm_dx_pos; FS.
now left.
Qed.


(* old version 
Lemma minorationenergy2: (*73p38-140909*)
  forall n,
   sqrt
     (squared_norm_dx xmin xmax (fst dX)
        (fun j : Z =>
          unh dX j (n+1) -
          unh dX j n)) <=     (* was convergence_error *)
   (sqrt 2 / sqrt (1 - Rsqr (CFL c dX)) * snd dX *
       sqrt (energy c (unh dX) dX xmin xmax n)). *)

Lemma minorationenergy2: (*73p38-140909*)
  forall n,
  0 < fst dX -> 0 < snd dX ->
     infinite_norm_dx (fst dX)
        (fun j : Z =>
          unh dX j (n+1) -
          unh dX j n) <=     (* was convergence_error *)
   (sqrt 2 / sqrt (2*eps - Rsqr eps) * snd dX *
       sqrt (energy c (unh dX) dX n)).
intros.
assert (snd dX <> 0) by auto with real.
assert (0 < 2*eps - Rsqr eps).
replace (2*eps - Rsqr eps) with (eps*(2-eps)) by (unfold Rsqr;ring). 
apply Rmult_lt_0_compat.
apply eps_lt.
apply Rplus_lt_reg_r with eps; ring_simplify.
apply Rlt_trans with 1.
apply eps_lt.
auto with real.
apply Rsqr_incr_0_var.
unfold infinite_norm_dx; rewrite Rsqr_sqrt.
repeat rewrite Rsqr_mult.
rewrite Rsqr_sqrt.
apply Rmult_le_reg_l with ((1 / 2 * (1 - Rsqr (CFL c dX)))/(Rsqr (snd dX))). 
unfold Rdiv; repeat apply Rmult_lt_0_compat.
auto with real.
auto with real.
apply Rplus_lt_reg_r with (Rsqr (CFL c dX)); ring_simplify.
rewrite <- Rsqr_1.
apply Rsqr_incrst_1.
apply Rle_lt_trans with (1-eps).
apply CFL_eps.
apply Rplus_lt_reg_r with (-1+eps); ring_simplify.
apply eps_lt.
unfold CFL.
unfold Rdiv; repeat apply Rmult_le_pos; auto with real.
auto with real.
auto with real.
apply Rle_trans with (1 / 2 * (1 - (CFL c dX)²) *
       infinite_squared_norm_dx (fst dX)
         (fun j : Z => 1 / snd dX * (unh dX j (n + 1) - unh dX j n))).
right.
unfold Rdiv; rewrite Rmult_assoc.
apply f_equal.
replace (/ (Rsqr (snd dX))) with (/(snd dX) * / (snd dX)).
rewrite <- infinite_squared_norm_dx_homogen; FS.
apply infinite_squared_norm_dx_eq; FS.
intros; unfold scalar_multiply_n.
ring.
unfold Rsqr; field; auto with real.
apply Rle_trans with (energy c (unh dX) dX n).
apply minorationenergy; trivial.
rewrite <- (Rmult_1_l (energy c (unh dX) dX n)) at 1.
rewrite <- Rmult_assoc.
apply Rmult_le_compat_r.
apply energy_pos; trivial.
rewrite Rsqr_div.
repeat rewrite Rsqr_sqrt.
apply Rle_trans with ((1 - (CFL c dX)²) / (2 * eps - eps²)).
2: right; field.
apply Rmult_le_reg_l with (2 * eps - eps²).
exact H2.
rewrite Rmult_1_r.
apply Rle_trans with ((1 - (CFL c dX)²)).
unfold CFL_condition_r in CFL_eps.
apply Rle_trans with (1-Rsqr (1-eps)).
unfold Rsqr; right; ring.
unfold Rminus; apply Rplus_le_compat_l.
apply Ropp_le_contravar.
apply Rsqr_incr_1.
exact CFL_eps.
unfold CFL.
unfold Rdiv; repeat apply Rmult_le_pos; auto with real.
apply Rplus_le_reg_l with eps; ring_simplify.
left; apply eps_lt.
right; field.
now apply Rgt_not_eq.
split.
now apply Rgt_not_eq.
apply Rgt_not_eq.
apply Rlt_gt; auto with real.
now left.
auto with real.
apply Rgt_not_eq.
now apply sqrt_lt_R0.
apply energy_pos.
exact H0.
exact H.
apply infinite_squared_norm_dx_pos; FS.
now left.
repeat apply Rmult_le_pos.
apply sqrt_pos.
left; apply Rinv_0_lt_compat.
now apply sqrt_lt_R0.
auto with real.
apply sqrt_pos.
Qed.

(** Energy is smaller than... *)
Lemma energy_majoration_aux: 
 forall n, (0 < n)%nat -> 0 < snd dX-> 0 < fst dX ->
  sqrt (energy c (unh dX) dX n)
   <= sqrt (energy c (unh dX) dX (n-1))
     +  (sqrt 2 / (2*sqrt (2*eps-Rsqr eps)))
           * (snd dX) * (infinite_norm_dx (fst dX) 
                               (fun j => fh j n)).
intros n Hn HdX2 HdX1.
assert (K:snd dX <> 0) by auto with real.
apply Rplus_le_reg_l with 
  (-sqrt (energy c (unh dX) dX (n - 1))); ring_simplify.
assert (0 <= sqrt (energy c (unh dX) dX n)).
apply sqrt_pos.
case H; intros H1; clear H.
apply Rmult_le_reg_l with
  (sqrt (energy c (unh dX) dX (n - 1)) +
     sqrt (energy c (unh dX) dX n)).
rewrite <- (Rplus_0_l 0).
apply Rle_lt_trans with (sqrt (energy c (unh dX) dX (n - 1)) + 0).
apply Rplus_le_compat_r.
apply sqrt_pos.
now apply Rplus_lt_compat_l.
apply Rle_trans with (energy c (unh dX) dX n
 - energy c (unh dX) dX (n - 1)).
right; apply trans_eq with
   (Rsqr(sqrt (energy c (unh dX) dX n))
  -Rsqr(sqrt (energy c (unh dX) dX (n - 1)))).
unfold Rsqr; ring.
repeat rewrite Rsqr_sqrt; try reflexivity.
apply energy_pos; trivial.
apply energy_pos; trivial.
rewrite energy_increases; trivial.
apply Rle_trans with
  (1 / 2 *
 ( (infinite_norm_dx (fst dX) 
       (fun j => unh dX j (n + 1) - unh dX j (n - 1)))
  * (infinite_norm_dx (fst dX) 
  (fun j : Z => fh j n)))).
apply Rmult_le_compat_l.
fourier.
apply Rle_trans with (1:=RRle_abs _ ).
apply Rle_trans with (infinite_norm_dx (fst dX) (fun j : Z => fh j n)
   * infinite_norm_dx (fst dX) (fun j : Z => unh dX j (n + 1) - unh dX j (n - 1))).
apply infinite_Cauchy_Schwarz_dx; FS.
now left.
right; ring.
repeat rewrite <- Rmult_assoc.
apply Rmult_le_compat_r.
apply sqrt_pos.
apply Rmult_le_reg_l with 2.
auto with real.
rewrite <- Rmult_assoc.
replace (2*(1/2)) with 1 by field.
rewrite Rmult_1_l.
apply Rle_trans with
 ((infinite_norm_dx  (fst dX)
     (fun j : Z => unh dX j n - unh dX j (n-1)))
  +  ((infinite_norm_dx (fst dX)
     (fun j : Z => unh dX j (n + 1) - unh dX j n)))).
apply Rle_trans with 
 (infinite_norm_dx (fst dX) (fun j : Z => (unh dX j n - unh dX j (n - 1)) +
                 (unh dX j (n + 1) - unh dX j n))).
2: apply infinite_squared_norm_dx_triang; FS.
unfold infinite_norm_dx; right; apply f_equal.
apply infinite_squared_norm_dx_eq; FS.
intros; ring.
now left.
apply Rle_trans with
  (sqrt 2 / sqrt (2*eps - Rsqr eps) * snd dX *
        sqrt (energy c (unh dX) dX (n-1))
 + sqrt 2 / sqrt (2*eps - Rsqr eps) * snd dX *
        sqrt (energy c (unh dX) dX n)).
apply Rplus_le_compat.
pattern n at 1; replace n with ((n-1)+1)%nat.
apply minorationenergy2; trivial.
omega.
apply minorationenergy2; trivial.
right; field.
apply Rgt_not_eq.
apply sqrt_lt_R0.
replace (2*eps-Rsqr eps) with (eps *(2-eps)) by (unfold Rsqr;ring).
apply Rmult_lt_0_compat.
apply eps_lt.
apply Rplus_lt_reg_r with eps; ring_simplify.
apply Rlt_trans with 1; try apply eps_lt; auto with real.
rewrite <- H1.
apply Rle_trans with 0.
rewrite Rplus_0_r.
rewrite <- Ropp_0.
apply Ropp_le_contravar.
apply sqrt_pos.
unfold Rdiv; repeat apply Rmult_le_pos.
apply sqrt_pos.
left; apply Rinv_0_lt_compat.
apply Rmult_lt_0_compat.
auto with real.
apply sqrt_lt_R0.
replace (2*eps-Rsqr eps) with (eps *(2-eps)) by (unfold Rsqr;ring).
apply Rmult_lt_0_compat.
apply eps_lt.
apply Rplus_lt_reg_r with eps; ring_simplify.
apply Rlt_trans with 1; try apply eps_lt; auto with real.
now left.
apply sqrt_pos.
Qed.



Lemma energy_majoration: 
 0 < snd dX-> 0 < fst dX ->
  forall t,
    let k:= (kt t (snd dX) -1)%nat in 
    sqrt (energy c (unh dX) dX k) 
       <= sqrt (energy c (unh dX) dX 0)
               +  (sqrt 2 / (2*sqrt (2*eps-Rsqr eps)))  * (snd dX) * sum_f 1 k 
                     (fun n => (infinite_norm_dx (fst dX) 
                               (fun j => fh j n))). 
intros.
induction k.
unfold sum_f; simpl.
ring_simplify.
apply Rle_refl.
apply Rle_trans with (1:= energy_majoration_aux (S k) (lt_O_Sn _)  H H0).
replace (S k -1)%nat with k by omega.
rewrite sum_f_S.
2: omega.
rewrite Rmult_plus_distr_l.
rewrite <- Rplus_assoc.
apply Rplus_le_compat_r.
exact IHk.
Qed.

End Stab.


